<?php

// Exit if accessed directly
if (!defined('ABSPATH')) {
    exit;
}

/**
 * Methods related to Custom Order Fields
 *
 * @class RP_WCCF_Order
 * @package WooCommerce Custom Fields
 * @author RightPress
 */
if (!class_exists('RP_WCCF_Order')) {

class RP_WCCF_Order
{

    /**
     * Constructor class
     *
     * @access public
     * @return void
     */
    public function __construct()
    {
        // Display checkout field values
        add_action('add_meta_boxes', array($this, 'add_meta_box_checkout'), 99, 2);

        // Display backend fields
        add_action('add_meta_boxes', array($this, 'add_meta_box_order'), 99, 2);

        // Hide item meta of type file
        add_filter('woocommerce_hidden_order_itemmeta', array($this, 'hide_files_from_meta'));

        // Download links for file upload fields
        add_action('woocommerce_after_order_itemmeta', array($this, 'file_download_links'), 10, 3);

        // Save order field data
        add_action('save_post', array($this, 'process_order_field_data'), 10, 2);

        // Frontend order field data display
        add_action('woocommerce_order_details_after_order_table', array($this, 'frontend_order_field_data_display'));

        // Frontend checkout field data display
        $position = apply_filters('wccf_checkout_field_data_display_position', 10);
        add_action('woocommerce_order_details_after_order_table', array($this, 'frontend_checkout_field_data_display'), $position);
    }

    /**
     * Add meta box for order fields
     *
     * @access public
     * @param string $post_type
     * @param object $post
     * @return void
     */
    public function add_meta_box_order($post_type, $post)
    {
        // Not product?
        if ($post_type !== 'shop_order') {
            return;
        }

        // Get fields to display
        $fields = RP_WCCF_Conditions::filter_by_conditions(RP_WCCF::get_fields('order'), 'order');

        // Add meta box if we have any fields to display
        if (!empty($fields)) {
            add_meta_box(
                'wccf_order',
                apply_filters('wccf_context_label', RP_WCCF::opt('alias_order'), 'order', 'backend'),
                array($this, 'render_wccf_order'),
                'shop_order',
                'normal',
                'high'
            );
        }
    }

    /**
     * Add meta box to display checkout field values
     *
     * @access public
     * @param string $post_type
     * @param object $post
     * @return void
     */
    public function add_meta_box_checkout($post_type, $post)
    {
        // Not product?
        if ($post_type !== 'shop_order') {
            return;
        }

        // Get checkout fields stored for this order
        $fields = get_post_meta($post->ID, '_wccf_checkout', true);

        // Add meta box if we have any data to display
        if (!empty($fields)) {
            add_meta_box(
                'wccf_checkout',
                apply_filters('wccf_context_label', RP_WCCF::opt('alias_checkout'), 'checkout', 'backend'),
                array($this, 'render_wccf_checkout'),
                'shop_order',
                'normal',
                'high'
            );
        }
    }

    /**
     * Render backend order fields
     *
     * @access public
     * @param object $post
     * @return void
     */
    public function render_wccf_order($post)
    {
        // Get fields to display
        $fields = RP_WCCF_Conditions::filter_by_conditions(RP_WCCF::get_fields('order'), 'order');

        // Get stored values
        $stored = get_post_meta($post->ID, '_wccf_order', true);
        $stored = $stored === '' ? array() : $stored;

        // Display list of fields
        RP_WCCF_FB::display_fields($fields, 'order', $stored);
    }

    /**
     * Render checkout field data
     *
     * @access public
     * @param object $post
     * @return void
     */
    public function render_wccf_checkout($post)
    {
        // Get checkout fields stored for this order
        $fields = get_post_meta($post->ID, '_wccf_checkout', true);

        // Display data
        if (!empty($fields)) {

            // Start container
            echo '<div class="view"><table cellspacing="0" class="display_meta"><tbody>';

            // Iterate over fields and display data
            foreach ($fields as $field) {

                // File download link
                if ($field['type'] === 'file') {
                    $prepend = '<tr><th>' . $field['label'] . ':</th><td><p>';
                    $append = '</p></td></tr>';
                    echo RP_WCCF_FB::file_download_link_html($post->ID, $field, $prepend, $append, 'wccf_checkout');
                }

                // Other field types
                else {

                    // Start table row and display label
                    echo '<tr><th>' . $field['label'] . ':</th><td><p>';

                    // Display values
                    if (!empty($field['option_labels'])) {
                        echo implode(', ', $field['option_labels']);
                    }
                    else {
                        echo is_array($field['value']) ? implode(', ', $field['value']) : $field['value'];
                    }

                    // End table row
                    echo '</p></td></tr>';
                }
            }

            // End container
            echo '</tbody></table></div>';
        }
    }

    /**
     * Hide item meta of type file
     *
     * @access public
     * @param array $keys
     * @return array
     */
    public function hide_files_from_meta($keys)
    {
        global $post;

        if ($post && $post->post_type === 'shop_order') {

            // Load order object
            $order = new WC_Order($post->ID);

            // Iterate over order items
            foreach ($order->get_items() as $order_item_key => $order_item) {

                // Get stored fields
                $meta = get_post_meta($order_item_key, '_wccf', true);

                // Iterate over stored fields
                if ($meta !== '') {
                    foreach ($meta as $field) {
                        if ($field['type'] === 'file') {
                            $keys[] = $field['key'];
                        }
                    }
                }
            }
        }

        return $keys;
    }

    /**
     * Render file download links for file fields
     *
     * @access public
     * @param int $item_id
     * @param array $item
     * @param object $product
     * @return void
     */
    public function file_download_links($item_id, $item, $product)
    {
        $links = '';

        // Get post meta
        $post_meta = get_post_meta($item_id, '_wccf', true);

        // Iterate over fields used in this order item
        if (!empty($post_meta)) {
            foreach ($post_meta as $field) {
                if ($field['type'] === 'file' && !empty($field['file'])) {
                    $prepend = '<tr><th>' . $field['label'] . ':</th><td><p>';
                    $append = '</p></td></tr>';
                    $links .= RP_WCCF_FB::file_download_link_html($item_id, $field, $prepend, $append);
                }
            }
        }

        // Display list of downloads
        if (!empty($links)) {
            echo '<div class="view"><table cellspacing="0" class="display_meta"><tbody>' . $links . '</tbody></table></div>';
        }
    }

    /**
     * Process order field data on order save action
     *
     * @access public
     * @param int $post_id
     * @param object $post
     * @return void
     */
    public function process_order_field_data($post_id, $post)
    {
        // Only process posts with type shop_order
        if ($post->post_type !== 'shop_order') {
            return;
        }

        // Load existing fields (if any)
        $existing = get_post_meta($post_id, '_wccf_order', true);
        $existing = $existing === '' ? array() : $existing;

        // Sanitize fields and save values
        if ($fields = RP_WCCF_FB::process_field_data('order', $existing)) {
            update_post_meta($post_id, '_wccf_order', $fields);
        }
    }

    /**
     * Frontend order field data display
     *
     * @access public
     * @param object $order
     * @return void
     */
    public function frontend_order_field_data_display($order)
    {
        $display = array();

        // Get order fields
        $fields = RP_WCCF::get_fields('order');

        // Get stored field data
        $stored = get_post_meta($order->id, '_wccf_order', true);
        $stored = empty($stored) ? array() : $stored;

        // Iterate over fields and check if at least one of them is public and has a value
        if (!empty($fields)) {
            foreach ($fields as $field) {
                if (!empty($field['public'])) {
                    foreach ($stored as $stored_field) {
                        if (preg_replace('/^wccf_/i', '', $stored_field['key']) === $field['key'] && isset($stored_field['value'])) {
                            $display[] = array_merge($stored_field, array(
                                'value_to_display' => RP_WCCF_FB::format_value_for_frontend_display($stored_field, 'wccf_order', $order->id)
                            ));
                        }
                    }
                }
            }
        }

        // Allow developers to hide or modify some fields
        $display = apply_filters('wccf_frontend_order_field_data', $display, $order->id);

        // Include template if we have at least one public field with value
        if (!empty($display)) {
            RP_WCCF::include_template('order/order-field-data', array(
                'order'     => $order,
                'fields'    => $display,
            ));
        }
    }

    /**
     * Frontend checkout field data display
     *
     * @access public
     * @param object $order
     * @return void
     */
    public function frontend_checkout_field_data_display($order)
    {
        // Get stored field data
        $stored = get_post_meta($order->id, '_wccf_checkout', true);

        // Only display if there's at least one field saved
        if ($stored !== '' && is_array($stored) && !empty($stored)) {

            $display = array();

            // Format values for display
            foreach ($stored as $stored_field) {
                $display[] = array_merge($stored_field, array(
                    'value_to_display' => RP_WCCF_FB::format_value_for_frontend_display($stored_field, 'wccf_checkout', $order->id)
                ));
            }

            // Allow developers to hide or modify some fields
            $display = apply_filters('wccf_frontend_checkout_field_data', $display, $order->id);

            // Include template
            if (!empty($display)) {
                RP_WCCF::include_template('order/checkout-field-data', array(
                    'order'     => $order,
                    'fields'    => $display,
                ));
            }
        }
    }

}

new RP_WCCF_Order();

}